"""parse XML files and build a space with objects
it aggregates this files and manage redefine and exists attributes
"""
from typing import Optional

from .i18n import _
from .xmlreflector import XMLReflector
from .utils import normalize_family
from .error import SpaceObjShallNotBeUpdated, DictConsistencyError
from .path import Path
from .config import Config

# RougailObjSpace's elements that shall be forced to the Redefinable type
FORCE_REDEFINABLES = ('family', 'follower', 'service', 'disknod', 'variables')
# RougailObjSpace's elements that shall be forced to the UnRedefinable type
FORCE_UNREDEFINABLES = ('value',)
# RougailObjSpace's elements that shall not be modify
UNREDEFINABLE = ('multi', 'type')
# RougailObjSpace's elements that did not created automaticly
FORCE_ELEMENTS = ('choice', 'property_', 'leadership', 'information')
# XML text are convert has name
FORCED_TEXT_ELTS_AS_NAME = ('choice', 'property', 'value', 'target')


# _____________________________________________________________________________
# special types definitions for the Object Space's internal representation
class RootRougailObject:  # pylint: disable=R0903
    """Root object
    """
    def __init__(self,
                 xmlfiles,
                 name=None,
                 ):
        if not isinstance(xmlfiles, list):
            xmlfiles = [xmlfiles]
        self.xmlfiles = xmlfiles
        if name:
            self.name = name


class Atom(RootRougailObject):  # pylint: disable=R0903
    """Atomic object (means can only define one time)
    """


class Redefinable(RootRougailObject):  # pylint: disable=R0903
    """Object that could be redefine
    """


class UnRedefinable(RootRougailObject):  # pylint: disable=R0903
    """Object that could not be redefine
    """


class ObjSpace:  # pylint: disable=R0903
    """
    Base object space
    """


class RougailObjSpace:
    """Rougail ObjectSpace is an object's reflexion of the XML elements
    """

    def __init__(self,
                 xmlreflector: XMLReflector,
                 ) -> None:
        self.space = ObjSpace()
        self.paths = Path()

        self.forced_text_elts_as_name = set(FORCED_TEXT_ELTS_AS_NAME)
        self.list_conditions = {}
        self.valid_enums = {}
        self.booleans_attributs = []

        self.make_object_space_classes(xmlreflector)

    def make_object_space_classes(self,
                                  xmlreflector: XMLReflector,
                                  ) -> None:
        """Create Rougail ObjectSpace class types from DDT file
        It enables us to create objects like:
        File(), Variable(), Ip(), Family(), Constraints()... and so on.
        """

        for dtd_elt in xmlreflector.dtd.iterelements():
            attrs = {}
            if dtd_elt.name in FORCE_REDEFINABLES:
                clstype = Redefinable
            elif not dtd_elt.attributes() and dtd_elt.name not in FORCE_UNREDEFINABLES:
                clstype = Atom
            else:
                clstype = UnRedefinable
            forced_text_elt = dtd_elt.type == 'mixed'
            for dtd_attr in dtd_elt.iterattributes():
                if set(dtd_attr.itervalues()) == {'True', 'False'}:
                    # it's a boolean
                    self.booleans_attributs.append(dtd_attr.name)
                if dtd_attr.default_value:
                    # set default value for this attribute
                    default_value = dtd_attr.default_value
                    if dtd_attr.name in self.booleans_attributs:
                        default_value = self.convert_boolean(default_value)
                    attrs[dtd_attr.name] = default_value
                if dtd_attr.name == 'redefine':
                    # has a redefine attribute, so it's a Redefinable object
                    clstype = Redefinable
                if dtd_attr.name == 'name' and forced_text_elt:
                    # child.text should be transform has a "name" attribute
                    forced_text_elt = False

            if forced_text_elt is True:
                self.forced_text_elts_as_name.add(dtd_elt.name)

            # create ObjectSpace object
            setattr(self, dtd_elt.name, type(dtd_elt.name.capitalize(), (clstype,), attrs))
        for elt in FORCE_ELEMENTS:
            setattr(self, elt, type(self._get_elt_name(elt), (RootRougailObject,), dict()))

    @staticmethod
    def _get_elt_name(elt) -> str:
        name = elt.capitalize()
        if name.endswith('_'):
            name = name[:-1]
        return name

    @staticmethod
    def display_xmlfiles(xmlfiles: list) -> str:
        """The function format xmlfiles informations to generate errors
        """
        if len(xmlfiles) == 1:
            return '"' + xmlfiles[0] + '"'
        return '"' + '", "'.join(xmlfiles[:-1]) + '"' + ' and ' + '"' + xmlfiles[-1] + '"'

    def xml_parse_document(self,
                           xmlfile,
                           document,
                           space,
                           namespace,
                           ):
        """Parses a Rougail XML file and populates the RougailObjSpace
        """
        if namespace in ['services', Config['variable_namespace']]:
            msg = _(f'Namespace name "{namespace}" is not allowed in "{xmlfile}"')
            raise DictConsistencyError(msg, 21)
        if not namespace:
            namespace = Config['variable_namespace']
        redefine_variables = []
        self._xml_parse(xmlfile,
                        document,
                        space,
                        namespace,
                        redefine_variables,
                        )

    def _xml_parse(self,  # pylint: disable=R0913
                   xmlfile,
                   document,
                   space,
                   namespace,
                   redefine_variables,
                   ) -> None:
        # var to check unique family name in a XML file
        family_names = []
        for child in document:
            if not isinstance(child.tag, str):
                # doesn't proceed the XML commentaries
                continue
            if child.tag == 'family':
                if child.attrib['name'] in family_names:
                    msg = _(f'Family "{child.attrib["name"]}" is set several times in "{xmlfile}"')
                    raise DictConsistencyError(msg, 44)
                family_names.append(child.attrib['name'])
            try:
                # variable objects creation
                variableobj = self.get_variableobj(xmlfile,
                                                   child,
                                                   space,
                                                   namespace,
                                                   redefine_variables,
                                                   )
            except SpaceObjShallNotBeUpdated:
                continue
            self.set_text(child,
                          variableobj,
                          )
            self.set_attributes(xmlfile,
                                child,
                                variableobj,
                                )
            self.remove(child,
                        variableobj,
                        redefine_variables,
                        )
            self.set_path(namespace,
                          document,
                          variableobj,
                          )
            self.add_to_tree_structure(variableobj,
                                       space,
                                       child,
                                       namespace,
                                       )
            if list(child) != []:
                self._xml_parse(xmlfile,
                                child,
                                variableobj,
                                namespace,
                                redefine_variables,
                                )

    def get_variableobj(self,
                        xmlfile: str,
                        child: list,
                        space,
                        namespace,
                        redefine_variables,
                        ):  # pylint: disable=R0913
        """
        retrieves or creates Rougail Object Subspace objects
        """
        obj = getattr(self, child.tag)
        name = self._get_name(child, namespace)
        if Redefinable in obj.__mro__:
            return self.create_or_update_redefinable_object(xmlfile,
                                                            child.attrib,
                                                            space,
                                                            child,
                                                            name,
                                                            namespace,
                                                            redefine_variables,
                                                            )
        if Atom in obj.__mro__:
            if child.tag in vars(space):
                # Atom instance has to be a singleton here
                # we do not re-create it, we reuse it
                return getattr(space, child.tag)
            return obj(xmlfile, name)
        if child.tag not in vars(space):
            setattr(space, child.tag, [])
        return obj(xmlfile, name)

    def _get_name(self,
                  child,
                  namespace: str,
                  ) -> Optional[str]:
        if child.tag == 'variables':
            return namespace
        if 'name' in child.attrib:
            return child.attrib['name']
        if child.text and child.tag in self.forced_text_elts_as_name:
            return child.text.strip()
        return None

    def create_or_update_redefinable_object(self,
                                            xmlfile,
                                            subspace,
                                            space,
                                            child,
                                            name,
                                            namespace,
                                            redefine_variables,
                                            ):  # pylint: disable=R0913
        """A redefinable object could be created or updated
        """
        existed_var = self.get_existed_obj(name,
                                           space,
                                           child,
                                           namespace,
                                           )
        if existed_var:
            # if redefine is set to object, default value is False
            # otherwise it's always a redefinable object
            default_redefine = child.tag in FORCE_REDEFINABLES
            redefine = self.convert_boolean(subspace.get('redefine', default_redefine))
            if redefine is True:
                if isinstance(existed_var, self.variable):  # pylint: disable=E1101
                    if namespace == Config['variable_namespace']:
                        redefine_variables.append(name)
                    else:
                        redefine_variables.append(space.path + '.' + name)
                existed_var.xmlfiles.append(xmlfile)
                return existed_var
            exists = self.convert_boolean(subspace.get('exists', True))
            if exists is False:
                raise SpaceObjShallNotBeUpdated()
            xmlfiles = self.display_xmlfiles(existed_var.xmlfiles)
            msg = _(f'"{child.tag}" named "{name}" cannot be re-created in "{xmlfile}", '
                    f'already defined in {xmlfiles}')
            raise DictConsistencyError(msg, 45)
        # object deos not exists
        exists = self.convert_boolean(subspace.get('exists', False))
        if exists is True:
            # manage object only if already exists, so cancel
            raise SpaceObjShallNotBeUpdated()
        redefine = self.convert_boolean(subspace.get('redefine', False))
        if redefine is True:
            # cannot redefine an inexistant object
            msg = _(f'Redefined object in "{xmlfile}": "{name}" does not exist yet')
            raise DictConsistencyError(msg, 46)
        if child.tag not in vars(space):
            setattr(space, child.tag, {})
        return getattr(self, child.tag)(xmlfile, name)

    def get_existed_obj(self,
                        name: str,
                        space: str,
                        child,
                        namespace: str,
                        ) -> None:
        """if an object exists, return it
        """
        if child.tag in ['variable', 'family']:
            name = normalize_family(name)
        if isinstance(space, self.family):  # pylint: disable=E1101
            if namespace != Config['variable_namespace']:
                name = space.path + '.' + name
            if not self.paths.path_is_defined(name):
                return None
            old_family_name = self.paths.get_variable_family_path(name)
            if space.path != old_family_name:
                xmlfiles = self.display_xmlfiles(space.xmlfiles)
                msg = _(f'Variable was previously create in family "{old_family_name}", '
                        f'now it is in "{space.path}" in {xmlfiles}')
                raise DictConsistencyError(msg, 47)
            return self.paths.get_variable(name)
        # it's not a family
        children = getattr(space, child.tag, {})
        if name in children:
            return children[name]
        return None

    @staticmethod
    def convert_boolean(value: str) -> bool:
        """Boolean coercion. The Rougail XML may contain srings like `True` or `False`
        """
        if isinstance(value, bool):
            return value
        if value == 'True':
            return True
        return False

    def set_text(self,
                 child,
                 variableobj,
                 ) -> None:
        """set text
        """
        if child.text is None or child.tag in self.forced_text_elts_as_name:
            return
        text = child.text.strip()
        if text:
            variableobj.text = text

    def set_attributes(self,
                       xmlfile,
                       child,
                       variableobj,
                       ):
        """ set attributes to an object
        """
        redefine = self.convert_boolean(child.attrib.get('redefine', False))
        if redefine and child.tag == 'variable':
            # delete old values
            has_value = hasattr(variableobj, 'value')
            if has_value and len(child) != 0:
                del variableobj.value
        for attr, val in child.attrib.items():
            if redefine and attr in UNREDEFINABLE:
                xmlfiles = self.display_xmlfiles(variableobj.xmlfiles[:-1])
                msg = _(f'cannot redefine attribute "{attr}" for variable "{child.attrib["name"]}"'
                        f' in "{xmlfile}", already defined in {xmlfiles}')
                raise DictConsistencyError(msg, 48)
            if attr in self.booleans_attributs:
                val = self.convert_boolean(val)
            if attr == 'name' and getattr(variableobj, 'name', None):
                # do not redefine name
                continue
            setattr(variableobj, attr, val)

    def remove(self,
               child,
               variableobj,
               redefine_variables,
               ):
        """Rougail object tree manipulations
        """
        if child.tag == 'variable':
            if child.attrib.get('remove_check', False):
                self.remove_check(variableobj.name)
            if child.attrib.get('remove_condition', False):
                self.remove_condition(variableobj.name)
            if child.attrib.get('remove_fill', False):
                self.remove_fill(variableobj.name)
        if child.tag == 'fill' and child.attrib['target'] in redefine_variables:
            self.remove_fill(child.attrib['target'])

    def remove_check(self, name):
        """Remove a check with a specified target
        """
        remove_checks = []
        for idx, check in enumerate(self.space.constraints.check):  # pylint: disable=E1101
            if check.target == name:
                remove_checks.append(idx)
        remove_checks.sort(reverse=True)
        for idx in remove_checks:
            self.space.constraints.check.pop(idx)  # pylint: disable=E1101

    def remove_condition(self,
                         name: str,
                         ) -> None:
        """Remove a condition with a specified source
        """
        remove_conditions = []
        for idx, condition in enumerate(self.space.constraints.condition):  # pylint: disable=E1101
            if condition.source == name:
                remove_conditions.append(idx)
        remove_conditions.sort(reverse=True)
        for idx in remove_conditions:
            del self.space.constraints.condition[idx]  # pylint: disable=E1101

    def remove_fill(self,
                    name: str,
                    ) -> None:
        """Remove a fill with a specified target
        """
        remove_fills = []
        for idx, fill in enumerate(self.space.constraints.fill):  # pylint: disable=E1101
            if fill.target == name:
                remove_fills.append(idx)
        remove_fills.sort(reverse=True)
        for idx in remove_fills:
            self.space.constraints.fill.pop(idx)  # pylint: disable=E1101

    def set_path(self,
                 namespace,
                 document,
                 variableobj,
                 ):
        """Fill self.paths attributes
        """
        if isinstance(variableobj, self.variable):  # pylint: disable=E1101
            family_name = normalize_family(document.attrib['name'])
            self.paths.add_variable(namespace,
                                    normalize_family(variableobj.name),
                                    namespace + '.' + family_name,
                                    document.attrib.get('dynamic') is not None,
                                    variableobj,
                                    )
        elif isinstance(variableobj, self.family):  # pylint: disable=E1101
            family_name = normalize_family(variableobj.name)
            if namespace != Config['variable_namespace']:
                family_name = namespace + '.' + family_name
            self.paths.add_family(namespace,
                                  family_name,
                                  variableobj,
                                  )

    @staticmethod
    def add_to_tree_structure(variableobj,
                              space,
                              child,
                              namespace: str,
                              ) -> None:
        """add a variable to the tree
        """
        variableobj.namespace = namespace
        if isinstance(variableobj, Redefinable):
            name = variableobj.name
            if child.tag in ['family', 'variable']:
                name = normalize_family(name)
            getattr(space, child.tag)[name] = variableobj
        elif isinstance(variableobj, UnRedefinable):
            getattr(space, child.tag).append(variableobj)
        else:
            setattr(space, child.tag, variableobj)
