"""load XML file from directory
"""
from typing import List
from os.path import join, isfile
from os import listdir

from lxml.etree import DTD, parse, XMLSyntaxError  # pylint: disable=E0611

from .i18n import _
from .error import DictConsistencyError


class XMLReflector:
    """Helper class for loading the Creole XML file,
    parsing it, validating against the Creole DTD,
    writing the xml result on the disk
    """
    def __init__(self):
        self.dtd = None

    def parse_dtd(self, dtdfilename):
        """Loads the Creole DTD

        :raises IOError: if the DTD is not found

        :param dtdfilename: the full filename of the Creole DTD
        """
        if not isfile(dtdfilename):
            raise IOError(_("no such DTD file: {}").format(dtdfilename))
        with open(dtdfilename, 'r') as dtdfd:
            self.dtd = DTD(dtdfd)

    def parse_xmlfile(self, xmlfile):
        """Parses and validates some Creole XML against the Creole DTD

        :returns: the root element tree object
        """
        try:
            document = parse(xmlfile)
        except XMLSyntaxError as err:
            raise DictConsistencyError(_(f'{xmlfile} is not an XML file: {err}'), 52) from err
        if not self.dtd.validate(document):
            dtd_error = self.dtd.error_log.filter_from_errors()[0]
            msg = _(f'"{xmlfile}" not a valid XML file: {dtd_error}')
            raise DictConsistencyError(msg, 43)
        return document.getroot()

    @staticmethod
    def load_xml_from_folders(xmlfolders: List[str]):
        """Loads all the XML files located in the xmlfolders' list

        :param xmlfolders: list of full folder's name
        """
        for xmlfolder in xmlfolders:
            filenames = [join(xmlfolder, filename) for filename in listdir(xmlfolder) if \
                         filename.endswith('.xml')]
            filenames.sort()
            for xmlfile in filenames:
                yield xmlfile
